from dataclasses import dataclass

from .. import endpoints
from ..session import Session
from ..exceptions import NetworkError
from .generics import TimeWindowDay

class Component:
    """Represents a component. Contains its parameters. Remember to call Component.update to populate values."""
    component_id: str
    display_name: str
    display_category: str
    standard_name: str
    type: str
    sub_type: str
    time_windows_view: list[TimeWindowDay]
    picture_url: str

    parameters: list['Parameter']

    def __init__(self, facility_id: int, component_id: str, session: Session):
        self.facility_id = facility_id
        self.component_id = component_id
        self.session = session

    @classmethod
    def from_overview_data(cls, facility_id: int, session: Session, obj: dict) -> 'Component':
        component = cls(facility_id, obj.get("componentId"), session)
        component.display_name = obj.get("displayName")
        component.display_category = obj.get("displayCategory")
        component.standard_name = obj.get("standardName")
        component.type = obj.get("type")
        component.sub_type = obj.get("subType")
        return component


    def __str__(self):
        return f'Component([Facility {self.facility_id}] -> {self.component_id})'

    async def update(self) -> list['Parameter']:
        res = await self.session.request("get", endpoints.COMPONENT.format(self.session.user_id, self.facility_id, self.component_id))
        self.component_id = res.get('componentId')
        self.display_name = res.get('displayName')
        self.display_category = res.get('displayCategory')
        self.standard_name = res.get('standardName')
        self.type = res.get('type')
        self.sub_type = res.get('subType')
        if res.get('timeWindowsView'):
            self.time_windows_view = TimeWindowDay.from_list(res['timeWindowsView'])

        #  TODO: Find endpoint that gives all parameters
        topview = res.get('topView')

        parameters = dict()
        if topview:
            self.picture_url = topview.get('pictureUrl')
            if 'pictureParams' in topview:
                parameters |= topview.get('pictureParams')
            if 'infoParams' in topview:
                parameters |= topview.get('infoParams')
            if 'configParams' in topview:
                parameters |= topview.get('configParams')
        if 'stateView' in res:
            parameters |= {i['name']: i for i in res.get('stateView')}
        if 'setupView' in res:
            parameters |= {i['name']: i for i in res.get('setupView')}

        self.parameters = Parameter.from_list(parameters.values(), self.session, self.facility_id)
        return self.parameters


@dataclass
class Parameter:
    session: Session
    facility_id: int

    id: str
    display_name: str
    name: str
    editable: bool
    parameter_type: str
    unit: str
    value: str
    min_val: str
    max_val: str
    string_list_key_values: dict[str, str]

    @classmethod
    def from_dict(cls, obj, session: Session, facility_id: int):
        parameter_id = obj["id"]
        display_name = obj.get("displayName")
        name = obj.get("name")
        editable = obj.get("editable")
        parameter_type =  obj.get("parameterType")
        unit = obj.get("unit")
        value =  obj.get("value")
        min_val = obj.get("minVal")
        max_val = obj.get("maxVal")
        string_list_key_values = obj.get("stringListKeyValues")

        return cls(session, facility_id, parameter_id, display_name, name, editable, parameter_type, unit, value, min_val, max_val, string_list_key_values)

    @classmethod
    def from_list(cls, obj, session: Session, facility_id: int):
        return [cls.from_dict(i, session, facility_id) for i in obj]

    @property
    def display_value(self) -> str:
        if self.string_list_key_values:
            return self.string_list_key_values[str(self.value)]
        if self.unit:
            return f'{self.value} {self.unit}'
        return str(self.value)

    async def set_value(self, value):
        """Returns None if value is the same"""
        try:
            return await self.session.request('put',
                                              endpoints.SET_PARAMETER.format(self.session.user_id, self.facility_id, self.id),
                                              json={"value": str(value)}
                                              )
        except NetworkError as e:
            if e.status == 304: # unchanged
                return None


"""Fröling connect API Wrapper

This library is an unofficial API wrapper for the Fröling Web Portal (https://connect-web.froeling.com/).
As for now, this wrapper is read only. Altering settings is not implemented yet. It supports reading statistics from
your devices and reading notifications.

Github and documentation: https://https://github.com/Layf21/froeling-connect.py
"""

from .client import Froeling
from .datamodels import Facility, Component, Parameter, UserData, NotificationOverview, NotificationDetails, Address



from dataclasses import dataclass
from .generics import Address


@dataclass(frozen=True)
class UserData:
    email: str
    salutation: str
    firstname: str
    surname: str
    address: Address
    user_id: int
    lang: str
    role: str
    active: bool
    pictureUrl: str
    facilityCount: int

    @staticmethod
    def from_dict(obj: dict):
        email = obj['userData'].get("email")
        salutation = obj['userData'].get("salutation")
        firstname = obj['userData'].get("firstname")
        surname = obj['userData'].get("surname")
        address = Address.from_dict(obj['userData'].get("address"))
        user_id = obj['userData'].get("userId")
        lang = obj.get("lang")
        role = obj.get("role")
        active = obj.get("active")
        pictureUrl = obj.get("pictureUrl")
        facilityCount = obj.get("facilityCount")
        return UserData(email, salutation, firstname, surname, address, user_id, lang, role, active, pictureUrl,
                        facilityCount)



"""Manages authentication and error handling"""

from aiohttp import ClientSession
import json
import base64
import logging

from . import endpoints, exceptions


#headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36', 'Referer': 'https://connect-web.froeling.com/'}
# Seems like user-agent and referer are not required

class Session:
    user_id: int = None
    token: str = None
    def __init__(self,
                 username: str=None, password: str=None, token: str=None,
                 auto_reauth: bool=False,
                 token_callback=None,
                 lang: str = 'en',
                 logger: logging.Logger=None,
                 clientsession: ClientSession=None
                 ):
        assert token or (username and password), "Set either token or username and password."
        assert not (auto_reauth and not (username and password)), "Set username and password to use auto_reauth."

        self.clientsession = clientsession or ClientSession()
        self._headers = {'Accept-Language': lang }
        self.username = username
        self.password = password
        self.auto_reauth = auto_reauth
        self.token_callback = token_callback


        if token:
            self.set_token(token)

        self._logger = logger or logging.getLogger(__name__)
        self._reauth_previous = False # Did the previous request result in renewing the token?

    async def close(self):
        await self.clientsession.close()

    def set_token(self, token: str):
        """Sets the token used in Authorization and updates/sets user-id
        :param token The bearer token"""
        self._headers['Authorization'] = token
        try:
            self.user_id = json.loads(base64.b64decode(token.split('.')[1] + "==").decode("utf-8"))['userId']
        except:
            raise ValueError("Token is in an invalid format.")
        if self.token_callback and self.token:  # Only run when overwriting existing token
            self.token = token
            self.token_callback(token)
        else:
            self.token = token

    async def login(self) -> dict:
        """Gets a token using username and password
        :return: Json sent by server (includes userdata)"""
        data = {'osType': 'web', 'username': self.username, 'password': self.password}
        async with await self.clientsession.post(endpoints.LOGIN, json=data) as res:
            if res.status != 200:
                raise exceptions.AuthenticationError(f'Server returned {res.status}: "{await res.text()}"')
            self.set_token(res.headers['Authorization'])
            userdata = await res.json()
        self._logger.debug("Logged in with username and password.")
        return userdata

    async def request(self, method, url, headers=None, **kwargs):
        """

        :param method:
        :param url:
        :param headers: Additional headers used in the request
        :param kwargs:
        """
        self._logger.debug(f'Sent %s: %s', method.upper(), url)
        request_headers = self._headers
        if headers:
            request_headers |= headers

        try:
            async with await self.clientsession.request(method, url, headers=request_headers, **kwargs) as res:
                if 299 >= res.status >= 200:
                    r = await res.text()
                    self._logger.debug('Got %s', r)
                    self._reauth_previous = False
                    return await res.json()

                if res.status == 401:
                    if self.auto_reauth:
                        if self._reauth_previous:
                            raise exceptions.AuthenticationError("Reauth did not work.", await res.text())
                        self._logger.info('Error %s, renewing token...', await res.text())
                        await self.login()
                        self._logger.info('Reauthorized.')
                        self._reauth_previous = True
                        return await self.request(method, url, **kwargs)
                    else:
                        self._logger.error("Request unauthorized")
                        raise exceptions.AuthenticationError("Request not authorized: ", await res.text())
                else:
                    error_data = await res.text()
                    raise exceptions.NetworkError("Unexpected return code", status=res.status, url=res.url, res=error_data)

        except json.decoder.JSONDecodeError as e:
            raise exceptions.ParsingError(e.msg, e.doc, e.pos, url)



"""API Endpoints"""

LOGIN = 'https://connect-api.froeling.com/connect/v1.0/resources/login'
"""post data: {"username": username, "password": password}"""

USER = 'https://connect-api.froeling.com/connect/v1.0/resources/service/user/{}'
"""1: user_id"""

FACILITY = 'https://connect-api.froeling.com/connect/v1.0/resources/service/user/{}/facility'
"""1: user_id
facility_ids = res[*]["facilityId"]"""

OVERVIEW = 'https://connect-api.froeling.com/fcs/v1.0/resources/user/{}/facility/{}/overview'
"""1: user_id 2: facility_id"""

COMPONENT_LIST = 'https://connect-api.froeling.com/fcs/v1.0/resources/user/{}/facility/{}/componentList'
"""1: user_id 2: facility_id"""

COMPONENT = 'https://connect-api.froeling.com/fcs/v1.0/resources/user/{}/facility/{}/component/{}'
"""1: user_id  2: facility_id  3: component_id"""

NOTIFICATION_COUNT = 'https://connect-api.froeling.com/connect/v1.0/resources/service/user/{}/notification/count'
"""1: user_id"""

NOTIFICATION_LIST = 'https://connect-api.froeling.com/connect/v1.0/resources/service/user/{}/notification'
"""1: user_id"""

NOTIFICATION = 'https://connect-api.froeling.com/connect/v1.0/resources/service/user/{}/notification/{}'
"""1: user_id  2: notification_id"""

SET_PARAMETER = 'https://connect-api.froeling.com/fcs/v1.0/resources/user/{}/facility/{}/parameter/{}'
"""1: user_id  2: facility_id  3: parameter_id"""



"""Provides the main API Class"""
import logging
from aiohttp import ClientSession

from . import datamodels
from . import endpoints
from .session import Session


class Froeling:
    """The Froeling class provides access to the Fröling API."""

    # cached data (does not change often)
    _userdata: datamodels.UserData
    _facilities: dict[int, datamodels.Facility] = {}


    async def __aenter__(self):
        if not self.session.token:
            await self.login()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.close()

    def __init__(self, username: str = None, password: str = None, token: str = None, auto_reauth: bool = False,
                 token_callback=None, language: str = 'en', logger: logging.Logger = None, clientsession: ClientSession = None):
        """Initialize a :class:`Froeling` instance.
        Either username and password or a token is required.
                :param username: The email you use to log into your Fröling account.
                :param password: Your Fröling password (not required when using token).
                :param token: A valid token (not required when using username and password).
                :param auto_reauth: Automatically fetches a new token if the current one expires (requires password and username).
                :param max_retries: How often to retry a request if the request failed.
                :param token_callback: A function that is called when the token gets renewed (useful for saving the token).
                :param clientsession: When provided, network communication will go through this aiohttp session."""

        self.session = Session(username, password, token, auto_reauth, token_callback, language, logger, clientsession)
        self._logger = logger or logging.getLogger(__name__)

    async def login(self) -> datamodels.UserData:
        data = await self.session.login()
        self._userdata = datamodels.UserData.from_dict(data)
        return self._userdata

    async def close(self):
        await self.session.close()

    @property
    def user_id(self):
        return self.session.user_id

    @property
    def token(self):
        return self.session.token

    async def _get_userdata(self) -> datamodels.UserData:
        res = await self.session.request("get", endpoints.USER.format(self.session.user_id))
        return datamodels.UserData.from_dict(res)
    async def get_userdata(self):
        """Gets userdata (cached)"""
        if not self._userdata:
            self._userdata = await self._get_userdata()
        return self._userdata


    async def _get_facilities(self) -> list[datamodels.Facility]:
        """Gets all facilities connected with the account and stores them in this.facilities."""
        res = await self.session.request("get", endpoints.FACILITY.format(self.session.user_id))
        return datamodels.Facility.from_list(res, self.session)

    async def get_facilities(self) -> list[datamodels.Facility]:
        if not self._facilities:
            facilities = await self._get_facilities()
            self._facilities = {f.facility_id: f for f in facilities}
        return list(self._facilities.values())

    async def get_facility(self, facility_id) -> datamodels.Facility:
        if facility_id not in self._facilities:
            await self.get_facilities()
        assert facility_id in self._facilities, f"Facility with id {facility_id} not found."
        return self._facilities[facility_id]


    async def get_notification_count(self) -> int:
        """Gets unread notification count"""
        return (await self.session.request("get", endpoints.NOTIFICATION_COUNT.format(self.session.user_id)))["unreadNotifications"]

    async def get_notifications(self) -> list[datamodels.NotificationOverview]:
        res = await self.session.request("get", endpoints.NOTIFICATION_LIST.format(self.session.user_id))
        return [datamodels.NotificationOverview(n, self.session) for n in res]

    async def get_notification(self, notification_id: int) -> datamodels.NotificationDetails:
        res = await self.session.request("get", endpoints.NOTIFICATION.format(self.session.user_id, notification_id))
        return datamodels.NotificationDetails.from_dict(res)

    def get_component(self, facility_id: int, component_id: str):
        return datamodels.Component(facility_id, component_id, self.session)



class AuthenticationError(BaseException):
    pass


class NetworkError(BaseException):
    def __init__(self, msg, status, url, res):
        super().__init__(f'{msg}: Status: {status}, url: {url}\nResult: {res}')
        self.status = status
        self.url = url


class ParsingError(BaseException):
    def __init__(self, msg, doc, pos, url):
        lineno = doc.count('\n', 0, pos) + 1
        colno = pos - doc.rfind('\n', 0, pos)
        errmsg = 'Error while parsing API response while fetching %s. %s: line %d column %d (char %d)\n%s' % (url, msg, lineno, colno, pos, doc)
        super().__init__(self, errmsg)
        self.msg = msg
        self.doc = doc
        self.pos = pos
        self.url = url
        self.lineno = lineno
        self.colno = colno



from dataclasses import dataclass
import datetime

from .. import endpoints

class NotificationOverview:
    """Stores basic data of a notification."""
    id: int
    subject: str
    unread: bool
    date: datetime.date
    error_id: int
    type: str
    """Known Values: "ERROR", "INFO", "WARNING", "ALARM" """
    facility_id: int
    facility_name: str

    details: 'NotificationDetails'

    def __init__(self, data, session):
        self.session = session
        self.set_data(data)

    def set_data(self, data):
        self.data = data

        self.id = data.get('id')
        self.subject = data.get('subject')
        self.unread = data.get('unread')
        self.date = datetime.datetime.fromisoformat(data.get('notificationDate'))
        self.error_id = data.get('errorId')
        self.type = data.get('notificationType')
        self.facility_id = data.get('facilityId')
        self.facility_name = data.get('facilityName')

    """Gets additional information about this notification."""
    async def info(self):
        res = await self.session.request("get", endpoints.NOTIFICATION.format(self.session.user_id, self.id))
        self.details = NotificationDetails.from_dict(res)
        return self.details


@dataclass
class NotificationDetails(NotificationOverview):
    """Stores all data of a notification"""
    body: str
    sms: bool
    mail: bool
    push: bool
    notification_submission_state_dto: list['NotificationSubmissionState']
    errorSolutions: list['NotificationErrorSolutions']

    @classmethod
    def from_dict(cls, obj):
        body = obj.get("body")
        sms = obj.get("sms")
        mail = obj.get("mail")
        push = obj.get("push")
        submission_state = None
        if "notificationSubmissionStateDto" in obj:
            submission_state = NotificationSubmissionState.from_list(obj["notificationSubmissionStateDto"])
        error_solutions = None
        if "errorSolutions" in obj:
            error_solutions = NotificationErrorSolutions.from_list(obj["errorSolutions"])
        notificationDetailsObject = cls(body, sms, mail, push, submission_state, error_solutions)
        notificationDetailsObject.set_data(obj)
        return notificationDetailsObject

@dataclass
class NotificationSubmissionState:
    id: int
    recipient: str
    type: str
    submitted_to: str
    submission_result: str


    @classmethod
    def from_dict(cls, obj: dict) -> 'NotificationSubmissionState':
        _id = obj.get('id')
        recipient = obj.get('recipient')
        type = obj.get('type')
        """Known values: "EMAIL", "TOKEN" """
        submitted_to = obj.get('submittedTo')
        submission_result = obj.get('submissionResult')

        return NotificationSubmissionState(_id, recipient, type, submitted_to, submission_result)

    @classmethod
    def from_list(cls, obj: list[dict]):
        return [cls.from_dict(i) for i in obj]


@dataclass
class NotificationErrorSolutions:
    error_reason: str
    error_solution: str

    @classmethod
    def from_list(cls, obj: list[dict]) -> list['NotificationErrorSolutions']:
        return [cls(i['errorReason'], i['errorSolution']) for i in obj]


from dataclasses import dataclass

from ..session import Session
from .. import endpoints
from .generics import Address
from .component import Component, Parameter

@dataclass(frozen=True)
class Facility:
    session: Session
    facility_id: int
    equipmentNumber: int
    status: str
    name: str
    address: Address
    owner: str
    role: str
    favorite: bool
    allowMessages: bool
    subscribedNotifications: bool
    pictureUrl: str
    protocol3200Info: dict[str, str]
    hoursSinceLastMaintenance: int
    operationHours: int
    facilityGeneration: str

    @staticmethod
    def from_dict(obj: dict, session: Session) -> 'Facility':
        facility_id = obj.get("facilityId")
        equipmentNumber = obj.get("equipmentNumber")
        status = obj.get("status")
        name = obj.get("name")
        address = Address.from_dict(obj.get("address"))
        owner = obj.get("owner")
        role = obj.get("role")
        favorite = obj.get("favorite")
        allowMessages = obj.get("allowMessages")
        subscribedNotifications = obj.get("subscribedNotifications")
        pictureUrl = obj.get("pictureUrl")
        protocol3200Info = obj.get("protocol3200Info")
        hoursSinceLastMaintenance = int(protocol3200Info.get("hoursSinceLastMaintenance"))
        operationHours = int(protocol3200Info.get("operationHours"))


        facilityGeneration = obj.get("facilityGeneration")
        return Facility(session, facility_id, equipmentNumber, status, name, address, owner, role, favorite, allowMessages,
                        subscribedNotifications, pictureUrl, protocol3200Info, hoursSinceLastMaintenance, operationHours, facilityGeneration)

    @staticmethod
    def from_list(obj: list, session: Session):
        return [Facility.from_dict(i, session) for i in obj]

    async def get_components(self) -> list[Component]:
        res = await self.session.request("get", endpoints.COMPONENT_LIST.format(self.session.user_id, self.facility_id))
        return [Component.from_overview_data(self.facility_id, self.session, i) for i in res]

    def get_component(self, component_id: str):
        return Component(self.facility_id, component_id, self.session)




from enum import Enum
from dataclasses import dataclass

@dataclass(frozen=True)
class Address:
    street: str
    zip: int
    city: str
    country: str

    @staticmethod
    def from_dict(obj: dict) -> 'Address':
        street = obj.get("street")
        zip = obj.get("zip")
        city = obj.get("city")
        country = obj.get("country")
        return Address(street, zip, city, country)


class Weekday(Enum):
    MONDAY = "MONDAY"
    TUESDAY = "TUESDAY"
    WEDNESDAY = "WEDNESDAY"
    THURSDAY = "THURSDAY"
    FRIDAY = "FRIDAY"
    SATURDAY = "SATURDAY"
    SUNDAY = "SUNDAY"

@dataclass
class TimeWindowDay:
    id: int
    weekday: Weekday
    phases: list['TimeWindowPhase']

    @classmethod
    def from_dict(cls, obj: dict) -> 'TimeWindowDay':
        _id = obj["id"]
        weekday = Weekday(obj["weekDay"])
        phases = TimeWindowPhase.from_list(obj["phases"])

        return cls(_id, weekday, phases)

    @classmethod
    def from_list(cls, obj: list) -> list['TimeWindowDay']:
        return [cls.from_dict(i) for i in obj]

@dataclass
class TimeWindowPhase:
    start_hour: int
    start_minute: int
    end_hour: int
    end_minute: int

    @classmethod
    def from_dict(cls, obj: dict) -> 'TimeWindowPhase':
        sh = obj['startHour']
        sm = obj['startMinute']
        eh = obj['endHour']
        em = obj['endMinute']

        return cls(sh, sm, eh, em)

    @classmethod
    def from_list(cls, obj: list) -> list['TimeWindowPhase']:
        return [cls.from_dict(i) for i in obj]
